package de.hft.stuttgart.citydoctor2.connect.edge;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.hft.stuttgart.citydoctor2.datastructure.BoundarySurface;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.Lod;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;

/**
 * @author wewetzer
 * @version 1.0
 */
public class CppFeature extends CppReferenceHandlingBase {
	
	private native long createCppFeature(String type) throws OutOfMemoryError;

	private native void disposeCppFeature(long ptrToCppFeature) throws NativePointerCastException;

	// setting geometry and attributes
	private native boolean setGeometry(long ptrToFeature, int geomType, double[] pointArray, int[] pointIdMap,
			int[] polygonArray, int[] polygonIdMap) throws NativeException;

	private native boolean setLoD(long ptrToFeature, int LoD) throws NativeException;
	

	// gets the id of a corresponding C++ entity to a given Java entity id
	private native int getCppPointId(long ptrToFeature, int javaPointId) throws NativeException;

	private native int getCppPolygonId(long ptrToFeature, int javaPolygonId) throws NativeException;

	private native double[] getPoints(long ptrToFeature) throws NativeException;


	// debug native methods
	private native void printMesh(long ptrToFeature) throws NativeException;

	private native void writeMeshToInventor(long ptrToFeature, String fileName) throws NativeException;


	private String type;

	public CppFeature(String type) {
		this.type = type;
		createCppObject();
	}

	/**
	 * Creates an C++ feature object
	 * 
	 * @throws IllegalStateException      Thrown if the encapsulated "long-Pointer"
	 *                                    isn't 0
	 * @throws NativeOutOfMemoryException If there is no more memory on the native
	 *                                    side
	 * 
	 * @see {@link disposeCppObject}
	 * @see de.hft.stuttgart.citydoctor2.connect.edge.CppReferenceHandlingBase#createCppObject()
	 */
	private void createCppObject() {
		if (0L != ptrToCppObject) {
			throw new IllegalStateException("Can't create new C++ object, there is already one!");
		}
		// Maybe an exception is thrown
		ptrToCppObject = createCppFeature(type);
	}

	@Override
	public void disposeCppObject() throws NativePointerCastException {
		if (0L != ptrToCppObject) {
			disposeCppFeature(ptrToCppObject);
			ptrToCppObject = 0L;
		}
	}

	
	public void setGeometry(Geometry geometry) throws NativeException {
		checkNullPointer();
		CppGeometryData geometryData = createCppFeatureInput(geometry);
		setGeometry(ptrToCppObject, 0, geometryData.pointArr, geometryData.pointIdMap, geometryData.polygonArr, geometryData.polygonIdMap);
	}	
	
	public void setLoD(Lod lod) throws NativeException {
		checkNullPointer();
		int LoD = -1;
		if(lod == Lod.LOD0) {
			LoD = 0;
		}
		else if(lod == Lod.LOD1) {
			LoD = 1;
		}
		else if(lod == Lod.LOD2) {
			LoD = 2;
		}
		else if(lod == Lod.LOD3) {
			LoD = 3;
		}
		else if(lod == Lod.LOD4) {
			LoD = 4;
		}
		
		setLoD(ptrToCppObject, LoD);
	}	
	

	/**
	 * Returns the id of the corresponding C++ point to a given Java point id
	 * 
	 * @param javaPointId The given Java point id
	 * @return The corresponding C++ point id
	 * @throws IllegalArgumentException If the feature pointer is 0
	 * @throws NativeException          If an error occurred during the execution of
	 *                                  the native code
	 */
	public int getCppPointId(int javaPointId) throws NativeException {
		checkNullPointer();
		return this.getCppPointId(ptrToCppObject, javaPointId);
	}

	/**
	 * Returns the id of the corresponding C++ polygon to a given java polygon
	 * id
	 * 
	 * @param javaPolygonId The given Java polygon id
	 * @return The corresponding C++ polygon id
	 * @throws IllegalArgumentException If the feature pointer is 0
	 * @throws NativeException          If an error occurred during the execution of
	 *                                  the native code
	 */
	public int getCppPolygonId(int javaPolygonId) throws NativeException {
		checkNullPointer();
		return this.getCppPolygonId(ptrToCppObject, javaPolygonId);
	}


	public double[] getPoints() throws NativeException {
		checkNullPointer();
		return this.getPoints(ptrToCppObject);
	}

	/**
	 * Debug method, that prints the mesh
	 * 
	 * @throws IllegalArgumentException If the feature pointer is 0
	 * @throws NativeException          If an error occurred during the execution of
	 *                                  the native code
	 */
	public void printMesh() throws NativeException {
		checkNullPointer();
		this.printMesh(ptrToCppObject);
	}

	/**
	 * Writes the current stored level of detail 1 mesh in the file "filename.iv" as
	 * Open Inventor file.
	 * 
	 * @param fileName The name of the output file
	 * @throws IllegalArgumentException If the feature pointer is 0
	 * @throws NativeException          If an error occurred during the execution of
	 *                                  the native code n
	 */
	public void writeMeshToInventor(String fileName) throws NativeException {
		checkNullPointer();
		this.writeMeshToInventor(ptrToCppObject, fileName);
	}

	/**
	 * Writes the current stored level of detail 2 mesh in the file "filename.iv" as
	 * Open Inventor file.
	 * 
	 * @param fileName The name of the output file
	 * @throws IllegalArgumentException If the feature pointer is 0
	 * @throws NativeException          If an error occurred during the execution of
	 *                                  the native code n
	 */

	private void printGeomDataAsCppCode(CppGeometryData cppGeoData) {

		System.out.println("\n \n \n");

		double[] pnts = cppGeoData.pointArr;
		System.out.println("const size_t numCoords = " + pnts.length / 3 + ";");
		System.out.println("Coordinate3d * coords[" + pnts.length / 3 + "];");
		for (int i = 0; i < pnts.length; i += 3) {
			System.out.print("coords[ " + i / 3 + " ] = new Coordinate3d( ");
			System.out.println(pnts[i] + ", " + pnts[i + 1] + ", " + pnts[i + 2] + " );");
		}

		System.out.println("");
		int[] polys = cppGeoData.polygonArr;
		System.out.println("const size_t numPolys = " + cppGeoData.polygonIdMap.length + ";");
		System.out.println("const size_t sizePolys = " + polys.length + ";");
		System.out.print("int polyArr[] = {");
		for (int i = 0; i < polys.length - 1; ++i) {
			if (-1 == polys[i]) {
				System.out.println();
				System.out.print("   " + polys[i] + ", ");
			} else if (-2 == polys[i]) {
				System.out.print("   " + polys[i] + ", ");
			} else {
				System.out.print("   " + polys[i] + ", ");
			}
		}
		System.out.println("   " + polys[polys.length - 1] + " \n};\n");

		if (cppGeoData.boundarySurfaceTypeArr.length > 0) {
			System.out.println("BoundarySurfaceType types[numPolys] = {");
			for (int i = 0; i < cppGeoData.boundarySurfaceTypeArr.length; i++) {
				switch (cppGeoData.boundarySurfaceTypeArr[i]) {
				case (1): {
					System.out.println("   roofsurface,");
					break;
				}
				case (2): {
					System.out.println("   wallsurface,");
					break;
				}
				case (3): {
					System.out.println("   groundsurface,");
					break;
				}
				default: {
					System.out.println("   undefined,");
				}
				}
				;
			}
			System.out.println("};");
			System.out.println("return createHouse( coords, numCoords, polyArr, numPolys, sizePolys, types, true);");
		}
	}

	/**
	 * generates all the necessary geometric input for the CppFeature
	 * 
	 * @param geom
	 */
	private CppGeometryData createCppFeatureInput(Geometry geometry) {

		// ====================================================================
		// Vertices
		List<Vertex> vertices = geometry.getVertices();
		double[] pointArr = new double[vertices.size() * 3];
		int[] pointIdMap = new int[vertices.size()];

		Map<Vertex, Integer> vertexToIndexMap = new HashMap<>(vertices.size());
		for (int i = 0; i < vertices.size(); i++) {
			Vertex v = vertices.get(i);
			vertexToIndexMap.put(v, i);
			pointArr[i * 3 + 0] = v.getX();
			pointArr[i * 3 + 1] = v.getY();
			pointArr[i * 3 + 2] = v.getZ();
			pointIdMap[i] = i;
		}

		// ====================================================================
		// Polygons & BoundarySurfaceType

		List<Polygon> polygons = geometry.getPolygons();
		int numReferencedVertices = 0;
		int[] polygonIdMap = new int[polygons.size()];
		for (int i = 0; i < polygons.size(); i++) {
			Polygon p = polygons.get(i);
			polygonIdMap[i] = i;
			numReferencedVertices += p.getExteriorRing().getVertices().size();
			for (LinearRing lr : p.getInnerRings()) {
				numReferencedVertices += lr.getVertices().size();
			}
		}

		int[] polygonArr = new int[numReferencedVertices];
		int[] boundarySurfaceTypeArr = new int[polygons.size()];
		int polyCounter = 0;
		for (int polyIndex = 0; polyIndex < polygons.size(); polyIndex++) {
			Polygon p = polygons.get(polyIndex);
			// exterior ring = -1
			polygonArr[polyCounter++] = -1;
			for (int i = 0; i < p.getExteriorRing().getVertices().size() - 1; i++) {
				Vertex v = p.getExteriorRing().getVertices().get(i);
				int id = vertexToIndexMap.get(v);
				polygonArr[polyCounter++] = id;
			}
			for (LinearRing lr : p.getInnerRings()) {
				// interior ring = -2
				polygonArr[polyCounter++] = -2;
				for (int i = 0; i < lr.getVertices().size() - 1; i++) {
					Vertex v = lr.getVertices().get(i);
					int id = vertexToIndexMap.get(v);
					polygonArr[polyCounter++] = id;
				}
			}

			if (geometry.getLod() == Lod.LOD2) {
				BoundarySurface bs = p.getPartOfSurface();
				if (bs != null) {
					boundarySurfaceTypeArr[polyIndex] = bs.getType().ordinal();
				} else {
					boundarySurfaceTypeArr[polyIndex] = 999999;
				}
			}
		}
		return new CppGeometryData(pointArr, pointIdMap, polygonArr, polygonIdMap, boundarySurfaceTypeArr);
	}

	private class CppGeometryData {
		private double[] pointArr;
		private int[] pointIdMap;
		private int[] polygonArr;
		private int[] polygonIdMap;
		private int[] boundarySurfaceTypeArr;

		private CppGeometryData(double[] pointArr, int[] pointIdMap, int[] polygonArr, int[] polygonIdMap,
				int[] bfTypeArr) {
			this.pointArr = pointArr;
			this.pointIdMap = pointIdMap;
			this.polygonArr = polygonArr;
			this.polygonIdMap = polygonIdMap;
			this.boundarySurfaceTypeArr = bfTypeArr;
		}
	}

	public String getType() {
		return type;
	}

}
